@@ -140,7 +140,7 @@ const WindowTitle = new Lang.Class({
 
     _init: function(metaWindow) {
         this._metaWindow = metaWindow;
-        this.actor = new St.BoxLayout();
+        this.actor = new St.BoxLayout({ style_class: 'window-button-box' });
 
         let app = Shell.WindowTracker.get_default().get_window_app(metaWindow);
         this._icon = new St.Bin({ style_class: 'window-button-icon',
@@ -217,6 +217,9 @@ const WindowButton = new Lang.Class({
         this._switchWorkspaceId =
             global.window_manager.connect('switch-workspace',
                                           Lang.bind(this, this._updateVisibility));
+        this._workspaceChangedId =
+            this.metaWindow.connect('workspace-changed',
+                                    Lang.bind(this, this._updateVisibility));
         this._updateVisibility();
 
         this._notifyFocusId =
@@ -281,6 +284,7 @@ const WindowButton = new Lang.Class({
     },
 
     _onDestroy: function() {
+        this.metaWindow.disconnect(this._workspaceChangedId);
         global.window_manager.disconnect(this._switchWorkspaceId);
         global.display.disconnect(this._notifyFocusId);
         this._contextMenu.destroy();
@@ -388,7 +392,8 @@ const AppButton = new Lang.Class({
                                                x_align: St.Align.START });
         stack.add_actor(this._singleWindowTitle);
 
-        this._multiWindowTitle = new St.BoxLayout({ x_expand: true });
+        this._multiWindowTitle = new St.BoxLayout({ style_class: 'window-button-box',
+                                                    x_expand: true });
         stack.add_actor(this._multiWindowTitle);
 
         this._icon = new St.Bin({ style_class: 'window-button-icon',
@@ -663,7 +668,7 @@ const WorkspaceIndicator = new Lang.Class({
         this.actor.connect('scroll-event', Lang.bind(this, this._onScrollEvent));
         this._updateMenu();
 
-        this._settings = new Gio.Settings({ schema: 'org.gnome.desktop.wm.preferences' });
+        this._settings = new Gio.Settings({ schema_id: 'org.gnome.desktop.wm.preferences' });
         this._settingsChangedId = this._settings.connect('changed::workspace-names', Lang.bind(this, this._updateMenu));
     },
 
@@ -775,20 +780,6 @@ const WindowList = new Lang.Class({
                 let spacing = node.get_length('spacing');
                 this._windowList.layout_manager.spacing = spacing;
             }));
-        this._windowList.connect('notify::allocation', Lang.bind(this,
-            function() {
-                if (this._groupingMode != GroupingMode.AUTO || this._grouped)
-                    return;
-
-                let allocation = this._windowList.allocation;
-                let width = allocation.x2 - allocation.x1;
-                let [, natWidth] = this._windowList.get_preferred_width(-1);
-                if (width < natWidth) {
-                    this._grouped = true;
-                    Meta.later_add(Meta.LaterType.BEFORE_REDRAW,
-                                   Lang.bind(this, this._populateWindowList));
-                }
-            }));
         this._windowList.connect('scroll-event', Lang.bind(this, this._onScrollEvent));
 
 	let indicatorsBox = new St.BoxLayout({ x_align: Clutter.ActorAlign.END });
@@ -832,6 +823,10 @@ const WindowList = new Lang.Class({
                                   Lang.bind(this, this._onWorkspacesChanged));
         this._onWorkspacesChanged();
 
+        this._switchWorkspaceId =
+            global.window_manager.connect('switch-workspace',
+                                          Lang.bind(this, this._checkGrouping));
+
         this._overviewShowingId =
             Main.overview.connect('showing', Lang.bind(this, function() {
                 this.actor.hide();
@@ -870,6 +865,7 @@ const WindowList = new Lang.Class({
         this._groupingModeChangedId =
             this._settings.connect('changed::grouping-mode',
                                    Lang.bind(this, this._groupingModeChanged));
+        this._grouped = undefined;
         this._groupingModeChanged();
     },
 
@@ -898,17 +894,55 @@ const WindowList = new Lang.Class({
         children[active].activate();
     },
 
+    _getPreferredUngroupedWindowListWidth: function() {
+        if (this._windowList.get_n_children() == 0)
+            return this._windowList.get_preferred_width(-1)[1];
+
+        let children = this._windowList.get_children();
+        let [, childWidth] = children[0].get_preferred_width(-1);
+        let spacing = this._windowList.layout_manager.spacing;
+
+        let workspace = global.screen.get_active_workspace();
+        let nWindows = global.display.get_tab_list(Meta.TabList.NORMAL, workspace).length;
+
+        return nWindows * childWidth + (nWindows - 1) * spacing;
+    },
+
+    _getMaxWindowListWidth: function() {
+        let indicatorsBox = this._trayButton.actor.get_parent();
+        return this.actor.width - indicatorsBox.get_preferred_width(-1)[1];
+    },
+
     _groupingModeChanged: function() {
         this._groupingMode = this._settings.get_enum('grouping-mode');
-        this._grouped = this._groupingMode == GroupingMode.ALWAYS;
-        this._populateWindowList();
+
+        if (this._groupingMode == GroupingMode.AUTO) {
+            this._checkGrouping();
+        } else {
+            this._grouped = this._groupingMode == GroupingMode.ALWAYS;
+            this._populateWindowList();
+        }
+    },
+
+    _checkGrouping: function() {
+        if (this._groupingMode != GroupingMode.AUTO)
+            return;
+
+        let maxWidth = this._getMaxWindowListWidth();
+        let natWidth = this._getPreferredUngroupedWindowListWidth();
+
+        let grouped = (maxWidth < natWidth);
+        if (this._grouped !== grouped) {
+            this._grouped = grouped;
+            this._populateWindowList();
+        }
     },
 
     _populateWindowList: function() {
         this._windowList.destroy_all_children();
 
         if (!this._grouped) {
-            let windows = Meta.get_window_actors(global.screen).sort(
+            let windows = global.get_window_actors().sort(
                 function(w1, w2) {
                     return w1.metaWindow.get_stable_sequence() -
                            w2.metaWindow.get_stable_sequence();
@@ -980,6 +1014,9 @@ const WindowList = new Lang.Class({
         if (win.skip_taskbar)
             return;
 
+        if (!this._grouped)
+            this._checkGrouping();
+
         if (this._grouped)
             return;
 
@@ -997,14 +1034,14 @@ const WindowList = new Lang.Class({
     },
 
     _onWindowRemoved: function(ws, win) {
-        if (this._grouped) {
-            if (this._groupingMode == GroupingMode.AUTO) {
-                this._grouped = false;
-                this._populateWindowList();
-            }
+        if (this._grouped)
+            this._checkGrouping();
 
+        if (this._grouped)
             return;
-        }
+
+        if (win.get_compositor_private())
+            return; // not actually removed, just moved to another workspace
 
         let children = this._windowList.get_children();
         for (let i = 0; i < children.length; i++) {
@@ -1117,6 +1154,9 @@ const WindowList = new Lang.Class({
         global.screen.disconnect(this._nWorkspacesChangedId);
         this._nWorkspacesChangedId = 0;
 
+        global.window_manager.disconnect(this._switchWorkspaceId);
+        this._switchWorkspaceId = 0;
+
         Main.messageTray.actor.anchor_y = 0;
         Main.messageTray._notificationWidget.anchor_y = 0;
 
@@ -1127,7 +1167,7 @@ const WindowList = new Lang.Class({
 
         this._settings.disconnect(this._groupingModeChangedId);
 
-        let windows = Meta.get_window_actors(global.screen);
+        let windows = global.get_window_actors();
         for (let i = 0; i < windows.length; i++)
             windows[i].metaWindow.set_icon_geometry(null);
     }
